# Classification of in-vivo NMR spectra to predict the mutation status of glioma patients.
# Support vector machine (svm) and data binning and t-test based feature selection (Border: min 2 features, 
# max 2 % of dataset) is used to classify the mutation status.

# Set your current Workspace directory
setwd("your workspace directory")

# Install necessary packages
install.packages("xlsx")

# Import necessary libraries
library(readxl)
library(e1071)
library(ggplot2)
library(xlsx)

# Import dataset
tumor_spektren<-read.csv2("./tumor_spektren_unedited.csv",header=TRUE,sep=";",quote="\"",dec=".",fill=TRUE,comment.char="", row.names=1)

# Start preprocessing of dataset

# Remove first column of the dataset which contains information about the mutation status
tum_data<-tumor_spektren[,-1]

# Initialize region of choline signal
choline_start<-345
choline_end<-370
# Initialize vector to save the maximum value of choline
choline_max_value<-numeric(length(tum_data[,1]))
# Initialize vector to save the corresponding column of the maximum value of choline
choline_max_column<-numeric(length(tum_data[,1]))

# Search in each spectra for the maximum choline signal
for (j in 1:dim(tum_data)[1]){
  for (k in choline_start:choline_end){
    # Check if current value is bigger than before
    if (tum_data[j,k]>choline_max_value[j]){
      # Save the value of maximum choline signal
      choline_max_value[j]<-tum_data[j,k]
      # Save number of corresponding column
      choline_max_column[j]<-k
    }
  }
}


# Data alignment:
# Shift each spectrum according to maximum choline signal so that maximum value 
# of each spectra is in the same column.

# Save column number of left maximum choline signal of all spectra
left_max_choline<-min(choline_max_column)
# Save column number of right maximum choline signal of all spectra
right_max_choline<-max(choline_max_column)
# Calculate maximum shift range of choline signal
choline_shift_range<-right_max_choline-left_max_choline

# Calculate total number of columns after shifting
# Notice: Column number starts at Index 1 not at 0
column_number<-(length(tum_data[1,])-choline_shift_range)+1

# Initialize matrix for shifted spectra
tum_data_shifted<-matrix(nrow=length(tum_data[,1]),ncol=column_number)

# Execute shifting for each spectra
for (j in 1:length(tum_data[,1])){
  # Calculate for each spectra the column difference between the maximum choline signal and the left_max_choline signal
  current_column_diff<-choline_max_column[j] - left_max_choline
  
  # Shift each datapoint in the spectrum according to the column difference
  for (k in 1:(column_number-1)){
    tum_data_shifted[j,k]<-tum_data[j,current_column_diff +k]
  }
}


# Removal of residual-water signal and empty regions:
# Empty regions: value of columns 1 to 89 and 451 to 1014 is near to zero
# Residual-water signal: sharp peak in region of column 500
tum_data_processed<-tum_data_shifted[,c(90:450)]


# Data normalization:
# Initialize vector to save the sum of each spectrum
sum_edited_spectra<-numeric(length(tum_data_processed[,1]))
# Initialize matrix to save the normalized data
tum_data_norm<-matrix(nrow=length(tum_data_processed[,1]),ncol=length(tum_data_processed[1,]))

# Calculate the sum of each spectrum
for (j in 1:length(tum_data_processed[,1])){
  sum_edited_spectra[j]<-sum(tum_data_processed[j,])
}

# Execute data normalization for each spectra
for (j in 1:length(tum_data_processed[,1])){
  for (k in 1:length(tum_data_processed[1,])){
    # Normalize data set
    tum_data_norm[j,k]<-((tum_data_processed[j,k])/sum_edited_spectra[j])*1000000
  }
}


# log transformation:
# Use log2 for transformation of data
tum_data_norm_log<-log2(abs(tum_data_norm))

# Save finally preprocessed matrix for classification as a data.frame, because the later 
# used function predict need a data.frame object.
classify_tum_data<-data.frame(tum_data_norm_log)


# Data binning:
# Calculate for each 4 signals the mean of them.

# Value for step size for reducing
reduce_step<-4
# Save new size of columns after reducing
new_col_size<-trunc(dim(classify_tum_data)[2]/reduce_step)
# Initialize vector to save values of each four datapoints
extract_vec<-numeric(length=reduce_step)
# Initialize matrix to save reduced datapoints
reduced_tum_data<-matrix(nrow=dim(classify_tum_data)[1], ncol=new_col_size+1)

# Execute data binning
for(i in 1:dim(classify_tum_data)[1]){
  l = 1
  j = 1
  while(l<=new_col_size*reduce_step){
    for(k in 1:reduce_step){
      extract_vec[k]<-(classify_tum_data[i,l+k-1])
    }
    # Save mean value of four datapoints in new matrix
    reduced_tum_data[i,j]<-sum(extract_vec)/reduce_step
    l = l + reduce_step
    j = j + 1
  }
  # Calculate mean of remaining values
  remaining_values<-(new_col_size*reduce_step)+1
  reduced_tum_data[i,new_col_size+1]<-sum(classify_tum_data[i,remaining_values:dim(classify_tum_data)[2]])/length(classify_tum_data[i,remaining_values:dim(classify_tum_data)[2]])
}

# Save new matrix as data frame
classify_tum_data<-data.frame(reduced_tum_data)

# End of preprocessing


# Start classification:
# Use leave-one-out cross-validation and Support Vector Machine and data binning 
# and t-test based feature selection (Border: min 2 features, max 2 % of dataset)

# Save label of each spectrum in 1-dim. vector
cv_data_labels<-tumor_spektren[,1]
# Initialize matrix to save accuracy of all spectra
cv_acc_complete<-matrix(nrow=dim(classify_tum_data)[2], ncol=dim(classify_tum_data)[1])
# Initialize vector to save p-values
p_values<- numeric(dim(classify_tum_data)[2])
# Initialize vector to save label prediction
cv_percent_pred_test<-numeric(dim(classify_tum_data)[1])
# Initialize matrix to save classification results
results<-matrix(nrow=dim(classify_tum_data[1]), ncol=3)
# Initialize matrix to save comparison of original label and prediction label
results_label<-matrix(nrow=dim(classify_tum_data[1]), ncol=2)

# Initialize border for optimal number of features
# Lower Border
min_num_feat<-2
# Upper Border with maximum 2 % of dataset
max_num_feat<-0
percentage<-2

# Create 2x2 windows for plotting Number of Features against Accuracay
par(mfrow=c(2,2))
# Create pdf document to save plot
pdf("Results_SVM_First_Max_Acc_Border.pdf")

# Loop through each spectra and do leave-one-out cross-validation
for (i in 1:dim(classify_tum_data)[1]){
  
  # Divide dataset in train and test data
  # Create train data using all data without i-th row.
  cv_train_data<-classify_tum_data[-i,]
  # Create test data using only i-th row.
  cv_test_data<-classify_tum_data[i,]
  # Save labels of train data
  cv_train_labels<-cv_data_labels[-i]
  # Save labels of test data
  cv_test_labels<-cv_data_labels[i] 

  # t-test based feature selection:
  # Calculate p-values by using the t-test and sort the p-values in ascending order.
  # Determine optimal feature number by internal SVM cross-validation. Therefore
  # calculate the accuracy of the Support Vector Machine of the train data with
  # increasing number of features. Use then all features of the mean of first and
  # last time maximum accuracy as optimal feature number.
  
  # Calculate p-values using t-test
  for (column in 1:dim(cv_train_data)[2]){
    # Perform t-test of train data and save p-values
    p_values[column]<-t.test(cv_train_data[c(which(cv_train_labels == 0)), column],
                               cv_train_data[c(which(cv_train_labels != 0)), column])$p.value  
  }
  # Sort calculated p-values in ascending order
  p_ordered<-order(p_values)
  
  # Determine optimal number of features by inernal SVM cross-validation.
  # Initialize vector to save the accuracy of SVM
  cv_acc<-numeric(length(p_ordered))
  
  # Loop through columns ordered by p-values and perform SVM for train data
  for (column in 1:length(p_ordered)){
    # Increase number of columns sorted by p-values in each iteration and save the accuracy
    cv_acc[column]<-svm(cv_train_data[, p_ordered[1:column]], cv_train_labels, scale = TRUE, type = "C-classification", kernel = "linear",
                           cross = dim(cv_train_data)[1])$tot.accuracy
  }
  # Save accuracy of SVM in matrix
  cv_acc_complete[,i]<-cv_acc
  
  # Check for first time maximum accuracy
  # Initialize variable to save the value which has first maximum accuracy
  cv_acc_value_max_first<-0
  # Initialize variable to save the number of features with the first maximum accuracy
  count_feat_first_acc_max<-0
  # Initialize variable to save the number of featues with the last maximum accuracy
  count_feat_last_acc_max<-0
  
  # Save number of features which have for the first time maximum accuracy
  # Loop through each columns ordered by p-values
  for(j in 1:length(p_ordered)){
    # Check if current accuracy value is larger than the value of cv.acc.max_first
    if(cv_acc[j]>cv_acc_value_max_first) {
      # Save value which has for the first time maximum accuracy
      cv_acc_value_max_first<-cv_acc[j]
      # Save number of features for the first time maximal accuracy
      count_feat_first_acc_max<-j
    }
  }
  
  # Save number of features which have for the last time maximum accuracy  
  # Loop through columns ordered by p-values
  for(j in 1:length(p_ordered)){
    # Check if current accuracy value equals first maximum accuracy
    if(cv_acc[j]==cv_acc_value_max_first){
      # Save number of features for the last time maximal accuracy
      count_feat_last_acc_max<-j
    } 
  }
  
  # Calculate upper border with 2 % of train data
  max_num_feat<-round((length(p_ordered)*percentage)/100)
  
  # Determine average number of features for classification
  if(count_feat_first_acc_max<min_num_feat){
    count_feat_first_acc_max<-min_num_feat
  }
  else if(count_feat_first_acc_max>max_num_feat){
    count_feat_first_acc_max<-max_num_feat
  }
  
  # Print results of each iteration
  cat("for It=",i,"optimal feature number=",count_feat_first_acc_max ,"min no of feat=",count_feat_first_acc_max,"max no of feat=",count_feat_last_acc_max,"\n")
 
  # Save the feature selection ordered by p-value
  col_selection<-p_ordered[1:count_feat_first_acc_max ]
  # Print feature selection
  print(col_selection)
  
  
  # Determine Support Vector Machine model:
  # Use train data and the feature selection
  SVM<-svm(cv_train_data[,col_selection],cv_train_labels,kernel="linear", scale= TRUE, type = "C-classification") 

  # Predict label of test data by using trained SVM model and feature selection
  label_pred<-predict(SVM,cv_test_data[,col_selection])
  
  # Determine percental result of prediction
  # 100 means correct predicted, 0 means not correct predicted
  cv_percent_pred_test[i]<-100*sum(label_pred==cv_test_labels)
  # Print percental result of each iteration
  cat("test results for iteration",i,"=",cv_percent_pred_test[i],"\n") 
  
  # Save status of prediction
  # Save Prediction, Number of selected features and Columns
  # Save status of prediction
  status<-label_pred==cv_test_labels
  if(status==1){
    results[i,1]<-"TRUE"
  }
  else{
    results[i,1]<-"FALSE"
  }
  results[i,2]<-length(col_selection)
  results[i,3]<-paste(col_selection, collapse = ",")

  # Append column names to result matrix
  colnames(results)<-c("Prediction", "Number of Features of first maximum accuracy with border", "Columns")
  
  # # Plot Number of Features against Accuracy and label title and axis
  # plot(1:361, cv_acc, type="l", lwd=2, main=paste("Iteration",i), cex.main=1.5, 
  #      sub=paste("Number of features of first max accuracy with border=", count_feat_first_acc_max, "in column:", paste("", col_selection, collapse=",")),
  #      xlab="Number of Features", ylab="Accuracy", cex.lab = 1)
  # mtext(paste("Prediction: ", status), side=3)
  
  
  # Save results of comparison of original label and prediction label
  # Save original label of each spectrum
  results_label[i,1]<-cv_data_labels[i]
  
  # Check if label prediction is conform with original label and save the results
  if(label_pred==cv_test_labels){
    results_label[i,2]<-cv_test_labels
  }
  else if(label_pred!=cv_test_labels){
    results_label[i,2]<-"FALSE"
  }
  # Append column names to result matrix
  colnames(results_label)<-c("Original Labels", "Prediction")
}

# End of classification


# Save results in Excel file
# write.csv2(results, file="Results_SVM_First_Max_Acc_Border_DataBinning.csv", row.names = TRUE)
# Save results of comparison of original label and prediction label in Excel file
write.xlsx(results_label, file="Results_SVMBorder_DataBinning_1-4.xlsx", row.names=TRUE)

# End of code